@@ -109,19 +109,22 @@ error:
 }
 
 /* Prepend a vertex to the linked list "list" based on the equalities in "tab".
+ * Return isl_bool_true if the vertex was actually added and
+ * isl_bool_false otherwise.
+ * In particular, vertices with a lower-dimensional activity domain are
+ * not added to the list because they would not be included in any chamber.
+ * Return isl_bool_error on error.
  */
-static int add_vertex(struct isl_vertex_list **list,
+static isl_bool add_vertex(struct isl_vertex_list **list,
 	__isl_keep isl_basic_set *bset, struct isl_tab *tab)
 {
 	unsigned nvar;
-	unsigned nparam;
 	struct isl_vertex_list *v = NULL;
 
 	if (isl_tab_detect_implicit_equalities(tab) < 0)
-		return -1;
+		return isl_bool_error;
 
 	nvar = isl_basic_set_dim(bset, isl_dim_set);
-	nparam = isl_basic_set_dim(bset, isl_dim_param);
 
 	v = isl_calloc_type(tab->mat->ctx, struct isl_vertex_list);
 	if (!v)
@@ -140,13 +143,18 @@ static int add_vertex(struct isl_vertex_list **list,
 	if (!v->v.dom)
 		goto error;
 
+	if (v->v.dom->n_eq > 0) {
+		free_vertex_list(v);
+		return isl_bool_false;
+	}
+
 	v->next = *list;
 	*list = v;
 
-	return 0;
+	return isl_bool_true;
 error:
 	free_vertex_list(v);
-	return -1;
+	return isl_bool_error;
 }
 
 /* Compute the parametric vertices and the chamber decomposition
@@ -155,13 +163,10 @@ error:
 static __isl_give isl_vertices *vertices_empty(__isl_keep isl_basic_set *bset)
 {
 	isl_vertices *vertices;
-	unsigned nparam;
 
 	if (!bset)
 		return NULL;
 
-	nparam = isl_basic_set_dim(bset, isl_dim_param);
-
 	vertices = isl_calloc_type(bset->ctx, isl_vertices);
 	if (!vertices)
 		return NULL;
@@ -183,13 +188,10 @@ static __isl_give isl_vertices *vertices_empty(__isl_keep isl_basic_set *bset)
 static __isl_give isl_vertices *vertices_0D(__isl_keep isl_basic_set *bset)
 {
 	isl_vertices *vertices;
-	unsigned nparam;
 
 	if (!bset)
 		return NULL;
 
-	nparam = isl_basic_set_dim(bset, isl_dim_param);
-
 	vertices = isl_calloc_type(bset->ctx, isl_vertices);
 	if (!vertices)
 		return NULL;
@@ -270,7 +272,7 @@ static int is_independent(__isl_keep isl_mat *facets, int n, isl_int *f)
  *
  * If the constraint is (strictly) redundant in the tableau, selecting it would
  * result in an empty tableau, so it can't be selected.
- * If the set variable part of the constraint is not linearly indepedent
+ * If the set variable part of the constraint is not linearly independent
  * of the set variable parts of the already selected constraints,
  * the constraint cannot be selected.
  * If selecting the constraint results in an empty tableau, the constraint
@@ -474,9 +476,11 @@ __isl_give isl_vertices *isl_basic_set_compute_vertices(
 		}
 		if (selected == nvar) {
 			if (tab->n_dead == nvar) {
-				if (add_vertex(&list, bset, tab) < 0)
+				isl_bool added = add_vertex(&list, bset, tab);
+				if (added < 0)
 					goto error;
-				n_vertices++;
+				if (added)
+					n_vertices++;
 			}
 			init = 0;
 			continue;
@@ -574,14 +578,21 @@ error:
 
 /* Can "tab" be intersected with "bset" without resulting in
  * a lower-dimensional set.
+ * "bset" itself is assumed to be full-dimensional.
  */
-static int can_intersect(struct isl_tab *tab, __isl_keep isl_basic_set *bset)
+static isl_bool can_intersect(struct isl_tab *tab,
+	__isl_keep isl_basic_set *bset)
 {
 	int i;
 	struct isl_tab_undo *snap;
 
+	if (bset->n_eq > 0)
+		isl_die(isl_basic_set_get_ctx(bset), isl_error_internal,
+			"expecting full-dimensional input",
+			return isl_bool_error);
+
 	if (isl_tab_extend_cons(tab, bset->n_ineq) < 0)
-		return -1;
+		return isl_bool_error;
 
 	snap = isl_tab_snap(tab);
 
@@ -589,18 +600,18 @@ static int can_intersect(struct isl_tab *tab, __isl_keep isl_basic_set *bset)
 		if (isl_tab_ineq_type(tab, bset->ineq[i]) == isl_ineq_redundant)
 			continue;
 		if (isl_tab_add_ineq(tab, bset->ineq[i]) < 0)
-			return -1;
+			return isl_bool_error;
 	}
 
 	if (isl_tab_detect_implicit_equalities(tab) < 0)
-		return -1;
+		return isl_bool_error;
 	if (tab->n_dead) {
 		if (isl_tab_rollback(tab, snap) < 0)
-			return -1;
-		return 0;
+			return isl_bool_error;
+		return isl_bool_false;
 	}
 
-	return 1;
+	return isl_bool_true;
 }
 
 static int add_chamber(struct isl_chamber_list **list,
@@ -631,7 +642,7 @@ static int add_chamber(struct isl_chamber_list **list,
 	c->c.vertices = isl_alloc_array(tab->mat->ctx, int, n_vertices);
 	if (n_vertices && !c->c.vertices)
 		goto error;
-	c->c.dom = isl_basic_set_from_basic_map(isl_basic_map_copy(tab->bmap));
+	c->c.dom = isl_basic_set_copy(isl_tab_peek_bset(tab));
 	c->c.dom = isl_basic_set_set_rational(c->c.dom);
 	c->c.dom = isl_basic_set_cow(c->c.dom);
 	c->c.dom = isl_basic_set_update_from_tab(c->c.dom, tab);
@@ -708,7 +719,7 @@ static struct isl_facet_todo *create_todo(struct isl_tab *tab, int con)
 	if (!todo->constraint)
 		goto error;
 	isl_seq_neg(todo->constraint->el, tab->bmap->ineq[con], 1 + tab->n_var);
-	todo->bset = isl_basic_set_from_basic_map(isl_basic_map_copy(tab->bmap));
+	todo->bset = isl_basic_set_copy(isl_tab_peek_bset(tab));
 	todo->bset = isl_basic_set_set_rational(todo->bset);
 	todo->bset = isl_basic_set_cow(todo->bset);
 	todo->bset = isl_basic_set_update_from_tab(todo->bset, tab);
@@ -848,7 +859,7 @@ static int update_todo(struct isl_facet_todo *first, struct isl_tab *tab)
  * set.  As a minor optimization, we only consider those activity domains
  * that contain some arbitrary point.
  *
- * For each of interior facets of the chamber, we construct a todo item,
+ * For each of the interior facets of the chamber, we construct a todo item,
  * containing the facet and a constraint containing the other side of the facet,
  * for constructing the chamber on the other side.
  * While their are any todo items left, we pick a todo item and
@@ -1134,7 +1145,7 @@ static struct isl_tab *tab_for_shifted_cone(__isl_keep isl_basic_set *bset)
 
 	if (!bset)
 		return NULL;
-	tab = isl_tab_alloc(bset->ctx, bset->n_ineq + 1,
+	tab = isl_tab_alloc(bset->ctx, bset->n_eq + bset->n_ineq + 1,
 			    1 + isl_basic_set_total_dim(bset), 0);
 	if (!tab)
 		return NULL;
@@ -1206,13 +1217,15 @@ static __isl_give isl_vec *isl_basic_set_interior_point(
  *
  * We pick an interior point from one of the chambers and then make
  * all constraints that do not satisfy this point strict.
+ * For constraints that saturate the interior point, the sign
+ * of the first non-zero coefficient is used to determine which
+ * of the two (internal) constraints should be tightened.
  */
 int isl_vertices_foreach_disjoint_cell(__isl_keep isl_vertices *vertices,
 	int (*fn)(__isl_take isl_cell *cell, void *user), void *user)
 {
-	int i, j;
+	int i;
 	isl_vec *vec;
-	isl_int v;
 	isl_cell *cell;
 
 	if (!vertices)
@@ -1234,21 +1247,11 @@ int isl_vertices_foreach_disjoint_cell(__isl_keep isl_vertices *vertices,
 	if (!vec)
 		return -1;
 
-	isl_int_init(v);
-
 	for (i = 0; i < vertices->n_chambers; ++i) {
 		int r;
 		isl_basic_set *dom = isl_basic_set_copy(vertices->c[i].dom);
-		dom = isl_basic_set_cow(dom);
-		if (!dom)
-			goto error;
-		for (j = 0; i && j < dom->n_ineq; ++j) {
-			isl_seq_inner_product(vec->el, dom->ineq[j], vec->size,
-						&v);
-			if (!isl_int_is_neg(v))
-				continue;
-			isl_int_sub_ui(dom->ineq[j][0], dom->ineq[j][0], 1);
-		}
+		if (i)
+			dom = isl_basic_set_tighten_outward(dom, vec);
 		dom = isl_basic_set_set_integral(dom);
 		cell = isl_cell_alloc(isl_vertices_copy(vertices), dom, i);
 		if (!cell)
@@ -1258,97 +1261,95 @@ int isl_vertices_foreach_disjoint_cell(__isl_keep isl_vertices *vertices,
 			goto error;
 	}
 
-	isl_int_clear(v);
 	isl_vec_free(vec);
 
 	return 0;
 error:
-	isl_int_clear(v);
 	isl_vec_free(vec);
 	return -1;
 }
 
-int isl_vertices_foreach_cell(__isl_keep isl_vertices *vertices,
-	int (*fn)(__isl_take isl_cell *cell, void *user), void *user)
+isl_stat isl_vertices_foreach_cell(__isl_keep isl_vertices *vertices,
+	isl_stat (*fn)(__isl_take isl_cell *cell, void *user), void *user)
 {
 	int i;
 	isl_cell *cell;
 
 	if (!vertices)
-		return -1;
+		return isl_stat_error;
 
 	if (vertices->n_chambers == 0)
-		return 0;
+		return isl_stat_ok;
 
 	for (i = 0; i < vertices->n_chambers; ++i) {
-		int r;
+		isl_stat r;
 		isl_basic_set *dom = isl_basic_set_copy(vertices->c[i].dom);
 
 		cell = isl_cell_alloc(isl_vertices_copy(vertices), dom, i);
 		if (!cell)
-			return -1;
+			return isl_stat_error;
 
 		r = fn(cell, user);
 		if (r < 0)
-			return -1;
+			return isl_stat_error;
 	}
 
-	return 0;
+	return isl_stat_ok;
 }
 
-int isl_vertices_foreach_vertex(__isl_keep isl_vertices *vertices,
-	int (*fn)(__isl_take isl_vertex *vertex, void *user), void *user)
+isl_stat isl_vertices_foreach_vertex(__isl_keep isl_vertices *vertices,
+	isl_stat (*fn)(__isl_take isl_vertex *vertex, void *user), void *user)
 {
 	int i;
 	isl_vertex *vertex;
 
 	if (!vertices)
-		return -1;
+		return isl_stat_error;
 
 	if (vertices->n_vertices == 0)
-		return 0;
+		return isl_stat_ok;
 
 	for (i = 0; i < vertices->n_vertices; ++i) {
-		int r;
+		isl_stat r;
 
 		vertex = isl_vertex_alloc(isl_vertices_copy(vertices), i);
 		if (!vertex)
-			return -1;
+			return isl_stat_error;
 
 		r = fn(vertex, user);
 		if (r < 0)
-			return -1;
+			return isl_stat_error;
 	}
 
-	return 0;
+	return isl_stat_ok;
 }
 
-int isl_cell_foreach_vertex(__isl_keep isl_cell *cell,
-	int (*fn)(__isl_take isl_vertex *vertex, void *user), void *user)
+isl_stat isl_cell_foreach_vertex(__isl_keep isl_cell *cell,
+	isl_stat (*fn)(__isl_take isl_vertex *vertex, void *user), void *user)
 {
 	int i;
 	isl_vertex *vertex;
 
 	if (!cell)
-		return -1;
+		return isl_stat_error;
 
 	if (cell->n_vertices == 0)
-		return 0;
+		return isl_stat_ok;
 
 	for (i = 0; i < cell->n_vertices; ++i) {
-		int r;
+		isl_stat r;
 
 		vertex = isl_vertex_alloc(isl_vertices_copy(cell->vertices),
 					  cell->ids[i]);
 		if (!vertex)
-			return -1;
+			return isl_stat_error;
 
 		r = fn(vertex, user);
 		if (r < 0)
-			return -1;
+			return isl_stat_error;
 	}
 
-	return 0;
+	return isl_stat_ok;
 }
 
 isl_ctx *isl_vertices_get_ctx(__isl_keep isl_vertices *vertices)
